// Gui Checkers by Jonathan Kreuzer
// copyright 2005
// =================================================
// BOARD Representation. I changed to this representation after seeing code by Martin Fierz.
// All the other code I believe was made without seeing other checkers source.
/*              37  38  39  40
              32  33  34  35
                28  29  30  31
              23  24  25  26
                19  20  21  22
              14  15  16  17
                10  11  12  13
               5   6   7   8         */
// =================================================
#define PD 0
const int KV = 33;
const int KBoard[64] = { PD, PD, PD, PD, PD,
						 KV-3,  KV+0,  KV+0,  KV+0, PD,
						    KV+2,  KV+3,  KV+3,  KV+0,
						  KV+1,  KV+4,  KV+4,  KV+3, PD,
						    KV+4,  KV+6,  KV+6,  KV+1,
						  KV+1,  KV+6,  KV+6,  KV+4, PD,
						    KV+3,  KV+4,  KV+4,  KV+1,
						  KV+0,  KV+3,  KV+3,  KV+2, PD,
						    KV+0,  KV+0,  KV+0, KV-3};
const int EBoard[64] = { PD, PD, PD, PD, PD,
						 -3,  0,  0, 0, PD,
						   -2,  0,  0,  -3,
						 -3,  0,  0, -2, PD,
						   -2,  0,  0,  -3,
						 -3,  0,  0, -2, PD,
						   -2,  0,  0,  -3,
						 -3,  0,  0, -2, PD,
						    0,  0,  0, -3};

struct CBoard 
{
	void StartPosition ( int bResetRep );
	void Clear();
	void SetFlags();
	int EvaluateBoard (int ahead, int alpha, int beta);

	// Data
	unsigned __int64 HashKey;
	short nPSq, eval;
	char nWhite, nBlack, nBK, nWK, extra0, extra1;
	char SideToMove, Start;
	char Sqs[48];
};

CBoard g_Boardlist[MAX_SEARCHDEPTH + 1];
CBoard g_CBoard;
CBoard g_StartBoard;

int BoardLoc[66] = 
	{	0, 37, 0, 38, 0, 39, 0, 40,	32, 0, 33, 0, 34, 0, 35, 0, 0, 28, 0, 29, 0 ,30, 0, 31,	23, 0, 24, 0, 25, 0, 26, 0, 
		0, 19, 0, 20, 0, 21, 0, 22,	14, 0, 15, 0, 16, 0, 17, 0, 0, 10, 0, 11, 0, 12, 0, 13,	5, 0, 6, 0, 7, 0, 8, 0};
int BoardLocTo32[48] = { 0, 0, 0, 0, 0, 0, 1, 2, 3, 0, 4, 5, 6, 7, 8,  9, 10, 11, 0, 12, 13, 14, 15, 16, 17, 18, 19, 0, 20, 21, 22, 23, 24, 25, 26, 27, 0, 28, 29, 30, 31};
int BoardLoc32[32] = { 5, 6, 7, 8, 10, 11, 12, 13, 14, 15, 16, 17, 19, 20, 21, 22, 23, 24, 25, 26, 28, 29, 30, 31, 32, 33, 34, 35, 37, 38, 39, 40};
long FlipX ( int x )
{	int y = x&3;
	x ^= y;
	x += 3-y;
	return x;
}

// =================================================
//
//	 			TRANSPOSITION TABLE 
//
// =================================================
unsigned char g_ucAge = 0;

struct TEntry
	{
	// FUNCTIONS
	public:
	void inline Read ( unsigned long CheckSum, short alpha, short beta, int &bestmove, int &value, int depth, int ahead)
		{
		if (m_checksum == CheckSum ) //To be almost totally sure these are really the same position.  
			{
			int tempVal;
			// Get the Value if the search was deep enough, and bounds usable
			if (m_depth >= depth)
               {
			   if (abs (m_eval) > 1800) // This is a game ending value, must adjust it since it depends on the variable ahead
					{
					if (m_eval > 0) tempVal = m_eval - ahead + m_ahead;
					if (m_eval < 0) tempVal = m_eval + ahead - m_ahead;
					}
			   else tempVal = m_eval;
			   switch ( m_failtype )
					{
					case 0: value = tempVal;  // Exact value
						break; 
					case 1: if (tempVal <= alpha) value = tempVal; // Alpha Bound (check to make sure it's usuable)
						break;
					case 2: if (tempVal >= beta) value = tempVal; //  Beta Bound (check to make sure it's usuable)
						break;
					}
               }
			// Otherwise take the best move from Transposition Table                                                
			bestmove = m_bestmove; 
			}
		}

	void inline Write ( unsigned long CheckSum, short alpha, short beta, int &bestmove, int &value, int depth, int ahead)
		{
		if (m_age == g_ucAge && m_depth > depth && m_depth > 14) return; // Don't write over deeper entries from same search
		m_checksum = CheckSum;
		m_eval = value;
		m_ahead = ahead;
		m_depth = depth;
		m_bestmove = bestmove;
		m_age = g_ucAge;
		if (value <= alpha) m_failtype = 1;
		else if (value >= beta)  m_failtype = 2;
		else m_failtype = 0;
		}

	static void Create_HashFunction ();
	static unsigned __int64 HashBoard (const CBoard &Board);

	// DATA
	private:
		unsigned long m_checksum;
		short m_eval;
		short m_bestmove;
		char m_depth;
		char m_failtype, m_ahead;
		unsigned char m_age; 		
	};

// Declare and allocate the transposition table
TEntry *TTable = (TEntry *) malloc( sizeof (TEntry) * HASHTABLESIZE + 2);

// ------------------
//  Hash Board
//
//  HashBoard is called to get the hash value of the start board.
// ------------------
unsigned __int64 HashFunction [48][12], HashSTM;

void TEntry::Create_HashFunction ( )
	{
	for (int i = 0; i <48; i++)
        for (int x = 0; x < 9; x++)
			{HashFunction [i][x]  = rand() + (rand()*256) + (rand()*65536);
			 HashFunction [i][x] <<= 32;
			 HashFunction [i][x]  += rand() + (rand()*256) + (rand()*65536);
			}
	HashSTM = HashFunction[0][0];
	}       

unsigned __int64 TEntry::HashBoard (const CBoard &Board)
	{
	unsigned __int64 CheckSum = 0;

    for (int index = 5; index <= 40; index++)
		{
		int nPiece =  Board.Sqs [ index ];
		if (nPiece != EMPTY) CheckSum ^= HashFunction [ index ][ nPiece ];
        }
    if (Board.SideToMove == BLACK) 
		{
		CheckSum ^= HashSTM;
		}
	return CheckSum;
	}

// Check for a repeated board 
// (only checks every other move, because side to move must be the same)
__int64 RepNum[ MAX_GAMEMOVES ];

// with Hashvalue passed
int inline Repetition ( const __int64 HashKey, int nStart, int ahead)
    {
	int i;
	if (nStart > 0) i = nStart; else i = 0;
	
	if ((i&1) != (ahead&1)) i++;

	ahead-=2;
    for ( ; i < ahead; i+=2)
		 if (RepNum[i] == HashKey) return TRUE;
                
    return FALSE;
    }

void inline AddRepBoard ( const __int64 HashKey, int ahead )
	{
	RepNum[ ahead ] = HashKey;
    }

// =================================================
//
//	 			BOARD FUNCTIONS
//
// =================================================
void CBoard::Clear ( )
	{
	int i;
	for (i = 0; i< 48; i++) Sqs [i] = EMPTY;
	for (i = 0; i < 5; i++) Sqs [i] = INVALID;
	for (i = 41; i <  48; i++) Sqs [i] = INVALID;
	Sqs[9] = INVALID; Sqs[18] = INVALID; Sqs[27] = INVALID; Sqs[36] = INVALID; 
	SetFlags ();
	}

void CBoard::StartPosition ( int bResetRep )
   {
	Clear ();
	int i;
	for (i = 5; i <= 8; i++)  Sqs [i] = BPIECE;
    for (i = 10; i <= 17; i++) Sqs [i] = BPIECE;
    for (i = 28; i <= 35; i++) Sqs [i] = WPIECE;
    for (i = 37; i <= 40; i++) Sqs [i] = WPIECE;

	SideToMove = BLACK;
	if (bResetRep) g_nMoves = 0;

	SetFlags ();
	g_StartBoard = *this;
   }

void CBoard::SetFlags ( )
	{
	nWhite = 0; nBlack = 0; nBK = 0; nWK = 0;
	nPSq = 0;
	for (int i = 5; i < 41; i++) 
		{
		if ( (Sqs[i]&WHITE) && Sqs[i]<INVALID) {nWhite++;
			if (Sqs[i] == WKING) {nWK++; nPSq += KBoard[i];} else nPSq += EBoard[i];
			}
		if ( (Sqs[i]&BLACK) && Sqs[i]<INVALID) {nBlack++;
			if (Sqs[i] == BKING) {nBK++; nPSq -= KBoard[i];} else nPSq -= EBoard[i];
			}
		}
	HashKey = TEntry::HashBoard ( *this );
	}

// ------------------
//  Evaluate Board
//
//  I don't know much about checkers, this is the best I could do for an evaluation function
// ------------------
const int LAZY_EVAL_MARGIN = 62;

int CBoard::EvaluateBoard (int ahead, int alpha, int beta)
	{
	// Game is over?        
	if (nWhite == 0) return -2001 + ahead;
	if (nBlack == 0) return 2001 - ahead;
	int eval, square;

	// Number of pieces present. Scaled higher for less material
	if ((nWhite+nBlack) > 12 ) eval = (nWhite-nBlack) * 100;
		else eval = (nWhite-nBlack) * (160 - (nWhite + nBlack) * 5 );

	// Add in the Piece Sq table, also includes bonuses for kings
	eval+=nPSq;

	// surely winning advantage
	if (nWhite == 1 && nBlack >= 3 && nBlack < 8) eval = eval + (eval>>1); 
	if (nBlack == 1 && nWhite >= 3 && nWhite < 8) eval = eval + (eval>>1);	

	// Too far from the alpha beta window? Forget about the rest of the eval, it probably won't get value back within the window
	if (eval+LAZY_EVAL_MARGIN < alpha) return eval;
	if (eval-LAZY_EVAL_MARGIN > beta) return eval;

	// The losing side can make it tough to win the game by moving a King back and forth on the double corner squares.
	if (eval > 18)
		{if ( Sqs[8] == BKING || Sqs[13] == BKING ) {eval-=8; if (Sqs[8]==EMPTY  || Sqs[13]==EMPTY) eval-=10;}
		 if ( Sqs[32] == BKING || Sqs[37] == BKING) {eval-=8; if (Sqs[32]==EMPTY || Sqs[37]==EMPTY) eval-=10;}
		}
	if (eval < -18)
		{if ( Sqs[8] == WKING  || Sqs[13] == WKING) {eval+=8; if (Sqs[8]==EMPTY  || Sqs[13]==EMPTY) eval+=10;}
		 if ( Sqs[32] == WKING || Sqs[37] == WKING) {eval+=8; if (Sqs[32]==EMPTY || Sqs[37]==EMPTY) eval+=10;}
		}

	// Bonus for leaving pieces guarding the back row.
	static int BackRowGuard[8] = { 0, 4, 5, 13, 4, 20, 11, 25 };
	int nBackB = 0, nBackW = 0;
	if ( (nWK*2) < nWhite ) 
		{if (Sqs[6] == BPIECE) nBackB+=1;
		 if (Sqs[7] == BPIECE) nBackB+=2;
		 if (Sqs[8] == BPIECE) nBackB+=4;
		 eval -= BackRowGuard [ nBackB ];}
	if ( (nBK*2) < nBlack )
		{if (Sqs[37] == WPIECE) nBackW+=1;
		 if (Sqs[38] == WPIECE) nBackW+=2;
		 if (Sqs[39] == WPIECE) nBackW+=4;
		 eval += BackRowGuard [ nBackW ];}

	// Number of Active Kings
	int nAWK = nWK, nABK = nBK;
	// Kings trapped on back row
	if (Sqs [5] == WKING && Sqs[6]  == BPIECE) {eval-=22; nAWK--; if (Sqs[14] != EMPTY) eval+=7;}
	if (Sqs [10] == WKING && Sqs[14]!=EMPTY && Sqs[6]==BPIECE) {eval-=10; nAWK--;}
	if (Sqs [6] == WKING && Sqs[15] == EMPTY && Sqs[5] == BPIECE && Sqs[7] == BPIECE) {eval-=14; nAWK--;}
	if (Sqs [7] == WKING && Sqs[16] == EMPTY && Sqs[6] == BPIECE && Sqs[8] == BPIECE) {eval-=14; nAWK--;}
	if (Sqs [8] == WKING && Sqs[13] != EMPTY && Sqs[7]  == BPIECE && Sqs[17] == EMPTY) {eval-=16; nAWK--;}
	if (Sqs [37] == BKING && Sqs[32] != EMPTY && Sqs[38] == WPIECE && Sqs[28] == EMPTY) {eval+=16; nABK--;}
	if (Sqs [38] == BKING && Sqs[29] == EMPTY && Sqs[37] == WPIECE && Sqs[39] == WPIECE) {eval+=14; nABK--;}
	if (Sqs [39] == BKING && Sqs[30] == EMPTY && Sqs[40] == WPIECE && Sqs[38] == WPIECE) {eval+=14; nABK--;}
	if (Sqs [40] == BKING && Sqs[39] == WPIECE) {eval+=22; nABK--; if (Sqs[31] != EMPTY) eval-=7; }
	if (Sqs [35] == BKING && Sqs[39] == WPIECE && Sqs[31] != EMPTY) {eval+=10; nABK--;}

	// Reward checkers that will king on the next move
	int KSQ = 0;
	for (square = 32; square  <= 35; square++)
		{
		if (Sqs [ square ] == BPIECE)
			{if (Sqs [ square + 4] == EMPTY) {KSQ = square+4; eval-=(KBoard [square+4] - 5); break;}
			 if (Sqs [ square + 5] == EMPTY) {KSQ = square+5; eval-=(KBoard [square+5] - 5); break;}
			}
		}
	// Opponent has no Active Kings and backrow is still protected, so give a bonus
	if (nAWK > 0 && (square == 36 || KSQ == 40) && nABK == 0) 
		{eval+=20;
		 if (nAWK > 1) eval+=10;
		 if (BackRowGuard [ nBackW ] > 10) eval += 15;
		 if (BackRowGuard [ nBackW ] > 20) eval += 20;
		}
	// Now the same for the other color
	for (square = 13; square  >= 10; square--)
		{
		if (Sqs [ square ] == WPIECE)
			{if (Sqs [ square - 4] == EMPTY )  {KSQ = square-4; eval+=(KBoard [square-4]-5); break;}
			 if (Sqs [ square - 5] == EMPTY )  {KSQ = square-5; eval+=(KBoard [square-5]-5); break;}
			}
		}
	if (nABK > 0 && (square == 14 || KSQ == 5) && nAWK == 0)
		{eval-=20;
		 if (nABK > 1) eval+=10;
		 if (BackRowGuard [ nBackB ] > 10) eval -= 15;
		 if (BackRowGuard [ nBackB ] > 20) eval -= 20;
		}

	// lower score, if equal number of pieces, but one side has all kinged versus all but one kinged (or all kings)
	if (nWhite == nBlack ) 
		{
		if (eval > 0 && eval < 200  && nWK >= nBK && nBK >= nWhite-1) eval = (eval>>1) + (eval>>3);
		if (eval < 0 && eval > -200 && nBK >= nWK && nWK >= nBlack-1) eval = (eval>>1) + (eval>>3);
		}

	return eval;
	}