// =================================================
//
//	 			MOVE GENERATION 
//
// =================================================
struct SMove 
	{
	 unsigned long SrcDst;
	 char cPath[12];
	};

struct CMoveList
	{
	// FUNCTIONS
	void inline Clear( ) {nJumps = 0;  
						  nMoves = 0;
						  }
	void inline SetMove (SMove &Move, int src, int dst)
		{
		Move.SrcDst = (src)+(dst<<6) + (1<<12);
		}
	void inline AddJump (SMove &Move, int pathNum)
		{
		Move.cPath[pathNum] = 0;
		Moves [ ++nJumps ] = Move;
		nMoves = nJumps;
		}
	void inline AddMove (int src, int dst)
		{
		if (nJumps != 0) return; // Ignore non-jumps if a jump move is present
		Moves [ ++nMoves ].SrcDst = (src)+(dst<<6);
		return;
		}

	void AddSqDir ( char board[], int square, int &nSqMoves, int nPiece, int pathNum, const int DIR, const int OPP_PIECE );
	void CheckMoveDir ( char board[], int square, const int DIR, const int nOppPiece, const int bJumpsOnly );
	void FindMovesBlack ( char board[]);
	void FindMovesWhite ( char board[]);
	void FindJumpsBlack ( char board[]);
	void FindJumpsWhite ( char board[]);
	void inline FindSqJumpsBlack ( char board[], int square, int nPiece, int pathNum, int nJumpSq);
	void inline FindSqJumpsWhite ( char board[], int square, int nPiece, int pathNum, int nJumpSq);

	// DATA
	int nMoves, nJumps;
	SMove Moves[36];
	SMove m_JumpMove;
	};

CMoveList g_Movelist[ MAX_SEARCHDEPTH + 1];
SMove g_GameMoves [ MAX_GAMEMOVES ];

// ------------------------------ FUNCTIONS --------------------------------
// Check for a normal move in DIR, then if a normal move is not possible, check for a jump
void inline CMoveList::CheckMoveDir ( char board[], int square, const int DIR, const int nOppPiece, const int bJumpsOnly )
{
	if ( !bJumpsOnly && board[square + DIR ] == EMPTY ) AddMove (square , square + DIR); 
		else if ( (board[square + DIR ]&nOppPiece) && board[square + 2*DIR ] == EMPTY) 
			{// A jump is possible, call FindSqJumps to detect double/triple/etc. jumps
			 SetMove (m_JumpMove, square, square + 2*DIR);
			 if (nOppPiece == WPIECE) FindSqJumpsBlack ( board, square + 2*DIR, board[ square ], 0, square + DIR);
			 if (nOppPiece == BPIECE) FindSqJumpsWhite ( board, square + 2*DIR, board[ square ], 0, square + DIR);
			} 
}
// -------------------------------------------------
// Find the Moves available on board, and store them in Movelist
// -------------------------------------------------
void CMoveList::FindMovesBlack ( char board[])
   {
	Clear ( );
	// Now loop through the legals squares on the board looking for black pieces
    for (register int square = 40; square  >= 5; square--)
		{
        if ( (board[ square ]&BPIECE) ) // Found a black piece
			{
			CheckMoveDir ( board, square, 4, WPIECE, 0 );
			CheckMoveDir ( board, square, 5, WPIECE, 0 );

			 // If it's a king, it can move backwards too
			if (board[ square ] == BKING  ) {
				CheckMoveDir ( board, square, -5, WPIECE, 0 );
				CheckMoveDir ( board, square, -4, WPIECE, 0 );
				}
			}
		}
  }

// For WHITE, this goes through the same process as the above function does for black
void CMoveList::FindMovesWhite ( char board[])
	{
	Clear ( );
    for (register int square = 5; square  <= 40; square++)
		{
        if ( (board[ square ]&WPIECE) ) 
			{
			CheckMoveDir ( board, square, -5, BPIECE, 0 );
			CheckMoveDir ( board, square, -4, BPIECE, 0 );

			if (board[ square ] == WKING  ) {
				CheckMoveDir ( board, square, 4, BPIECE, 0 );
				CheckMoveDir ( board, square, 5, BPIECE, 0 );
				}
			}
		}
	}
// -------------------------------------------------
// These two functions only add jumps
// they are used in the quiescence search
// -------------------------------------------------
void CMoveList::FindJumpsBlack ( char board[])
   {
	Clear ( );
    for (register int square = 40; square  >= 5; square--)
		{
        if ( (board[ square ]&BPIECE) ) // Found a black piece
			{
			CheckMoveDir ( board, square, 4, WPIECE, 1 );
			CheckMoveDir ( board, square, 5, WPIECE, 1 );

			if (board[ square ] == BKING  ) {
				CheckMoveDir ( board, square, -5, WPIECE, 1 );
				CheckMoveDir ( board, square, -4, WPIECE, 1 );
				}
			}
		}
  }    

void CMoveList::FindJumpsWhite ( char board[])
	{
	Clear ( );
    for (register int square = 5; square  <= 40; square++)
		{
        if ( (board[ square ]&WPIECE) ) 
			{
			CheckMoveDir ( board, square, -5, BPIECE, 1 );
			CheckMoveDir ( board, square, -4, BPIECE, 1 );

			if (board[ square ] == WKING  ) {
				 CheckMoveDir ( board, square, 4, BPIECE, 1 );
				 CheckMoveDir ( board, square, 5, BPIECE, 1 );
				}
			}
		}
	}

// -------------
//  If a jump move was possible, we must make the jump then check to see if the same piece can jump again.
//  There might be multiple paths a piece can take on a double jump, these functions store each possible path as a move.
// -------------
void inline CMoveList::AddSqDir ( char board [], int square, int &nSqMoves, int nPiece, int pathNum, const int DIR, const int OPP_PIECE )
	{
	if ( (board[square + DIR ]&OPP_PIECE) && board[square + 2*DIR ] == EMPTY ) 
		{m_JumpMove.cPath[ pathNum ] = square + 2*DIR; 
		 nSqMoves++; 
		 if (OPP_PIECE == WPIECE) FindSqJumpsBlack ( board, square+2*DIR, nPiece, pathNum+1, square + DIR);
		 if (OPP_PIECE == BPIECE) FindSqJumpsWhite ( board, square+2*DIR, nPiece, pathNum+1, square + DIR);
		}
 	}

void CMoveList::FindSqJumpsBlack ( char board[], int square, int nPiece, int pathNum, int nJumpSq)
{
	int nSqMoves = 0;
	// Remove the jumped piece (until the end of this function), so we can't jump it again
	int nJumpPiece = board [ nJumpSq ];
	board [ nJumpSq ] = EMPTY;

	// Now see if a piece on this square has more jumps
	AddSqDir ( board, square, nSqMoves, nPiece, pathNum, 4, WPIECE );
	AddSqDir ( board, square, nSqMoves, nPiece, pathNum, 5, WPIECE );
	if ( nPiece == BKING) 
		{// If this piece is a king, it can also jump backwards	
		AddSqDir ( board, square, nSqMoves, nPiece, pathNum, -5, WPIECE );
		AddSqDir ( board, square, nSqMoves, nPiece, pathNum, -4, WPIECE );
		}
	if (nSqMoves == 0) AddJump (m_JumpMove, pathNum); // this is a leaf, so store the move
	// Put back the jumped piece
	board [ nJumpSq ] = nJumpPiece;
}

void CMoveList::FindSqJumpsWhite ( char board[], int square, int nPiece, int pathNum, int nJumpSq)
{
	int nSqMoves = 0;
	int nJumpPiece = board [ nJumpSq ];
	board [ nJumpSq ] = EMPTY;

	AddSqDir ( board, square, nSqMoves, nPiece, pathNum, -4, BPIECE );
	AddSqDir ( board, square, nSqMoves, nPiece, pathNum, -5, BPIECE );	
	if ( nPiece == WKING) 
		{
		AddSqDir ( board, square, nSqMoves, nPiece, pathNum, 5, BPIECE );
		AddSqDir ( board, square, nSqMoves, nPiece, pathNum, 4, BPIECE );	
		}
	if (nSqMoves == 0) AddJump (m_JumpMove, pathNum);
	board [ nJumpSq ] = nJumpPiece;
}
//
// MOVE EXECUTION
//
// ---------------------
//  Helper Functions for DoMove 
// ---------------------
void inline DoSingleJump (CBoard &board, int src, int dst, int nPiece)
{
	int jumpedSq = (dst + src) >> 1;                    
    if (board.Sqs [ jumpedSq ] == WPIECE ) {board.nWhite--; board.nPSq -= EBoard [jumpedSq];}
        else if (board.Sqs [jumpedSq ] == BPIECE) {board.nBlack--; board.nPSq += EBoard [jumpedSq];}
        else if (board.Sqs [jumpedSq ] == BKING)  {board.nBlack--; board.nBK--; board.nPSq += KBoard [jumpedSq];}
        else if (board.Sqs [jumpedSq ] == WKING)  {board.nWhite--; board.nWK--; board.nPSq -= KBoard [jumpedSq];}

	board.HashKey   ^= HashFunction [ src ][ nPiece  ] ^ HashFunction  [ dst ][ nPiece  ] ^ HashFunction [ jumpedSq ][ board.Sqs [ jumpedSq  ] ];
	// Update the board array
    board.Sqs [ dst ] = nPiece;
    board.Sqs [ jumpedSq  ] = EMPTY;
    board.Sqs [ src ] = EMPTY;
}

// This function will test if a checker needs to be upgraded to a king, and upgrade if necessary
void inline CheckKing (CBoard &board, int dst, int nPiece)
{
	if (dst <= 8 && nPiece == WPIECE) 
		{board.Sqs [ dst  ] = WKING;
		 board.nPSq += KBoard[dst];
		 board.HashKey  ^= HashFunction [ dst ][ nPiece  ] ^ HashFunction [ dst ][ WKING ];
		 board.nWK++;
		}
	if (dst >= 37 && nPiece == BPIECE) 
		{board.Sqs [ dst  ] = BKING;
		 board.nPSq -= KBoard[dst];
		 board.HashKey  ^= HashFunction [ dst ][ nPiece  ] ^ HashFunction [ dst ][ BKING ];
		 board.nBK++;
		}
}

void inline UpdateSqTable ( CBoard &board, int nPiece, int src, int dst )
{
	if (nPiece == BKING) board.nPSq -= KBoard[ dst ] - KBoard [ src ];
	else if (nPiece == WKING) board.nPSq += KBoard[ dst ] - KBoard [ src ];
	else if (nPiece == WPIECE) board.nPSq += EBoard[ dst ] - EBoard [ src ];
	else if (nPiece == BPIECE) board.nPSq -= EBoard[ dst ] - EBoard [ src ];
}

// ---------------------
//  Execute a Move 
// ---------------------
int DoMove (SMove &Move, CBoard &board, int nType)
    {
	unsigned int src = (Move.SrcDst&63);
	unsigned int dst = (Move.SrcDst>>6)&63;
	unsigned int bJump = (Move.SrcDst>>12);
	unsigned int nPiece = board.Sqs [src];

	board.SideToMove ^=3;
	board.HashKey ^= HashSTM; // Flip hash stm
		
	if (bJump == 0) 
		{
		// Update hash values
		board.HashKey   ^= HashFunction [ src ][ nPiece  ] ^ HashFunction  [ dst ][ nPiece  ];

		// Update the board array
		board.Sqs [ dst ] = board.Sqs [ src ];
		board.Sqs [ src ] = EMPTY;

		UpdateSqTable ( board, nPiece, src, dst );
		if (nPiece < 4)	{
			 CheckKing ( board, dst, nPiece );
			 return 1;}
		return 0;
		}

	DoSingleJump (board, src, dst, nPiece);
	CheckKing ( board, dst, nPiece );

	// Double Jump?
	if (Move.cPath [0] == 0) 
		{UpdateSqTable ( board, nPiece, src, dst );
		 return 1;
		}
	if (nType == HUMAN) return DOUBLEJUMP;

	for (int i = 0; i < 8; i++)
		{
		int nDst = Move.cPath [i];
		if (nDst == 0) break;

		DoSingleJump (board, dst, nDst, nPiece);
		dst = nDst;
		}

	UpdateSqTable ( board, nPiece, src, dst );
	CheckKing ( board, dst, nPiece );      
    return 1;
    }