#include "board.cpp"
#include "movegen.cpp"

// ===============================================
//
//					 SEARCH 
//
// ===============================================
// -------------------------------------------------
// Quiescence Search... Search all jumps, if there are no jumps, stop the search
// -------------------------------------------------
int QuiesceBoard ( int color, int ahead, int alpha, int beta )
	{
	int value;

	if (color == WHITE) 
		{g_Movelist[ahead].FindJumpsWhite( g_Boardlist[ahead-1].Sqs );}
    else 
		{g_Movelist[ahead].FindJumpsBlack( g_Boardlist[ahead-1].Sqs );}

	// No more jump moves, return evaluation
	if (g_Movelist[ahead].nJumps == 0 || (ahead+1) >= MAX_SEARCHDEPTH) 
		{
		nodes++;
		if (ahead > g_SelectiveDepth) g_SelectiveDepth = ahead-1;
		return g_Boardlist[ahead-1].EvaluateBoard ( ahead-1, alpha, beta); 
		}

	// There were jump moves, so we keep searching.
	for (int i = 1; i <= g_Movelist[ahead].nJumps; i++)
		{
		g_Boardlist[ahead] = g_Boardlist[ahead-1];
		DoMove(g_Movelist[ahead].Moves[i] , g_Boardlist[ahead], SEARCHED);
		nodes++;
		// Recursive Call
        value = QuiesceBoard( (color^3) , ahead+1 , alpha, beta);

		// Keep Track of Best Move and Alpha-Beta Prune
		if (color == WHITE && value > alpha )
			{
			alpha = value;
			if (alpha >= beta ) return beta;
			}
		if (color == BLACK && value < beta )
			{
			beta = value;
			if (alpha >= beta )  return alpha;
			}
		}

	if (color == WHITE) return alpha; else return beta;
	}

//
// Extensions. fracDepth unit = 1/32 of a ply
//
void inline DoExtensions ( int &nextDepth, int &fracDepth, int ahead, unsigned long nMove )
{
	if (g_Movelist[ahead].nJumps == 1) 	fracDepth+=20; // Only 1 move
	if (g_Movelist[ahead].nJumps == 2)  fracDepth+=4;  // Only 2 moves
	if ( nextDepth < 6)
		{
		// These extend when a checker is nearing it's king square
		unsigned int dst = (nMove>>6)&63;
		if ( dst > 27 && dst < 32 && g_Boardlist[ahead].Sqs[dst] == BPIECE && g_Boardlist[ahead].Sqs[dst+9] == EMPTY ) fracDepth+=17;
		if ( dst > 13 && dst < 18 && g_Boardlist[ahead].Sqs[dst] == WPIECE && g_Boardlist[ahead].Sqs[dst-9] == EMPTY ) fracDepth+=17;
		if ( (dst > 31 && g_Boardlist[ahead].Sqs[dst] == BPIECE ) || (dst < 14 && g_Boardlist[ahead].Sqs[dst] == WPIECE )) fracDepth+=8;
		}
	if (fracDepth >= 32) {
		nextDepth++;
		fracDepth -= 32;}		
}

// -------------------------------------------------
//  Alpha Beta Search with Transposition(Hash) Table
// -------------------------------------------------
short ABSearch(int color, int ahead, int depth, int fracDepth, short alpha, short beta, int &bestmove)
{
int i, value, nextbest, nM;
unsigned long indexTT, checksumTT;
short temp;

// Check to see if move time has run out every 100,000 nodes
if (nodes > nodes2 + 100000 ) 
	{
	if (*g_pbPlayNow) return TIMEOUT;
    endtime = clock ();
    nodes2 = nodes;
	// If time has run out, we allow running up to 2*Time if g_bEndHard == FALSE and we are still searching a depth
    if ( (endtime - starttime) > (CLOCKS_PER_SEC * fMaxSeconds) )
		if ((endtime - starttime) > (2* CLOCKS_PER_SEC * fMaxSeconds) || g_bEndHard == TRUE || g_SearchingMove == 0 || abs(g_SearchEval) > 1500 ) 
			return TIMEOUT;
    }

// Use Internal Iterative Deepening to set bestmove if there's no best move
if (bestmove == NONE && depth > 4 )	{
	ABSearch (color, ahead, depth-4, 0, alpha, beta, bestmove);
	}
// Find possible moves (and set a couple variables)
if (color == WHITE) 
		g_Movelist[ahead].FindMovesWhite(g_Boardlist[ahead-1].Sqs );
	else 
		g_Movelist[ahead].FindMovesBlack(g_Boardlist[ahead-1].Sqs );

// If you can't move, you lose the game
if (g_Movelist[ahead].nMoves == 0) 
	{if (color == WHITE) return -2001+ahead; 
		else  return 2001 - ahead;
	}

// Run through the g_Movelist, doing each move
	for (i = 0; i <= g_Movelist[ahead].nMoves; i++)
		{
		// if bestmove is set, try Best Move at i=0, and skip it later
		if (i == bestmove) continue;
		if (i == 0) 
			{if (bestmove == NONE) continue;
			 nM = bestmove; 
			}
		else nM = i;

		if (ahead == 1) 
			{
			 g_SearchingMove = i; 
			 if (i > bestmove) g_SearchingMove--;
			 if (color == WHITE) g_SearchEval = alpha; else g_SearchEval = beta;
			 if (SearchDepth > 11) RunningDisplay ( bestmove, 1 );
			}
			 
		// Reset the move board to the one it was called with, then do the move # nM in the Movelist
        g_Boardlist[ahead] = g_Boardlist[ahead-1];
        temp = DoMove(g_Movelist[ahead].Moves[nM] , g_Boardlist[ahead], SEARCHED);
		AddRepBoard ( g_Boardlist[ahead].HashKey, g_nMoves + ahead);
		nodes++;

        // If the game is over, return a gameover value now
        if (g_Boardlist[ahead].nWhite == 0 || g_Boardlist[ahead].nBlack == 0) 
			{
			bestmove = nM;
            return g_Boardlist[ahead].EvaluateBoard ( ahead, -100000, 100000);
            }

        // If this is the max depth, quiesce then evaluate the board position
		if (depth <= 1 || ahead >= MAX_SEARCHDEPTH )
			{
			value = QuiesceBoard ( (color^3) , ahead+1, alpha, beta );
	        }
		// If this is the reptition of a position that has occured already in the search, return a draw score
		else if (depth >=4 && ahead>1 && Repetition (g_Boardlist[ahead].HashKey, g_nMoves-24, g_nMoves+ahead))
			{
			value = 0;
			}
        //If this isn't the max depth continue to look ahead 
         else {	
				value = -9999; nextbest = NONE;
				int nextDepth = depth - 1;

				DoExtensions ( nextDepth, fracDepth, ahead, g_Movelist[ahead].Moves[nM].SrcDst);

			   	// First look up the this Position in the Transposition Table
                if (nextDepth > 1 && hashing == 1 )
					{
					indexTT = ((unsigned long)g_Boardlist[ahead].HashKey) % HASHTABLESIZE;
					checksumTT = (unsigned long)(g_Boardlist[ahead].HashKey>>32);
					TTable [ indexTT ].Read ( checksumTT, alpha, beta, nextbest, value, nextDepth, ahead);
					}

				// Then try Forward Pruning, by searching to a reduced depth (nextDepth-R) 
				// then we check if the value is at least nMargin outside the alpha-beta window
				// If it is, we then skip the full search.
				if (value == -9999 && nextDepth > 2)
					{
					const int nMargin = 32, R = 3;
					g_Boardlist[ahead].eval = g_Boardlist[ahead].EvaluateBoard (ahead, alpha-100, beta+100);
					if (color == WHITE && g_Boardlist[ahead].eval > beta && beta > -1500) 
						{
						value = ABSearch( (color^3), ahead+1, nextDepth-R, fracDepth, beta+nMargin, beta+nMargin+1,  nextbest);
						if (value <= beta+nMargin) value = -9999;
						}
					else if (color == BLACK && g_Boardlist[ahead].eval < alpha && alpha < 1500) 
						{
						value = ABSearch( (color^3), ahead+1, nextDepth-R, fracDepth, alpha-nMargin-1, alpha-nMargin,  nextbest);
						if (value >= alpha-nMargin) value = -9999;
						}
					}

                // If value wasn't set from the Transposition Table or Pruning, look ahead nextDepth with a recursive call
                if (value == -9999 || ahead < 3)
					{
					// call ABSearch with opposite color & ahead incremented
                    value = ABSearch( (color^3), ahead+1, nextDepth, fracDepth, alpha, beta,  nextbest);
                    if (value == TIMEOUT) return value;
                         
                    // Store for the search info for this position in the Transposition Table
                    if (nextDepth > 1 && hashing == 1) 
						{
						TTable [ indexTT ].Write ( checksumTT, alpha, beta, nextbest, value, nextDepth, ahead);
						}
                    }
			}
             
		  // Penalize moves at root that repeat positions, so hopefully the computer will always make progress if possible
		  if (ahead == 1 && abs(value)<1000)
			{
			 if ( Repetition ( g_Boardlist[ahead].HashKey, 0, g_nMoves+1) ) value = (value>>1) + (value>>2);
			 else if (temp == 0) {if (color == WHITE) value-=1; else value+=1;} // if moves are the same, prefer the non-repeatable move
			}

          // Keep Track of Best Move and Alpha-Beta Prune
          if (color==WHITE && value > alpha )
				{
				 bestmove = nM;
				 alpha = value;
				 if (alpha >= beta ) return beta;
				}
          if (color==BLACK && value < beta )
				{
				 bestmove = nM;
				 beta = value;
				 if (alpha >= beta )  return alpha;
				}
		} // end for
                  
if (color == WHITE) return alpha; else return beta;
}

// -------------------------------------------------
// The computer calculates a move then updates g_CBoard.
// -------------------------------------------------
int ComputerMove ( char cColor, CBoard &InBoard)
	{
	int LastEval = 0, Eval, nps = 0, nDoMove = NONE, bestmove = NONE;
    CBoard TempBoard;
	CMoveList Movelist;
	g_ucAge++; // TT

	// return if game is over
	if (InBoard.nBlack == 0 || InBoard.nWhite == 0) return 0;

	if (cColor == BLACK) Movelist.FindMovesBlack(InBoard.Sqs );
	if (cColor == WHITE) Movelist.FindMovesWhite(InBoard.Sqs );
	if (Movelist.nMoves == 0) return 0; // game over

	srand ( clock() );
	bestmove = rand() % Movelist.nMoves + 1;

	 // Initialize variables (node count, start time, search depth)
	 nodes = 0;
     nodes2 = 0;
     starttime = clock ();
	 endtime = starttime;
	 g_SelectiveDepth = 0;
	 if (g_MaxDepth < 4) SearchDepth = g_MaxDepth-2;
        else if (g_MaxDepth <= 8 ) SearchDepth = 2;
		else SearchDepth = 6;
     Eval = 0;

     // Iteratively deepen until until the computer runs out of time, or reaches g_MaxDepth ply
     while ( SearchDepth < g_MaxDepth && Eval!=TIMEOUT)
         {
         SearchDepth +=2;
                                 
         g_Boardlist[0] = InBoard;
		 g_Boardlist[0].HashKey = TEntry::HashBoard ( g_Boardlist[0] );
         Eval = ABSearch (cColor, 1, SearchDepth, 0, -4000, 4000,  bestmove);
		 if (bestmove!=NONE) nDoMove = bestmove;

         if (Eval!=TIMEOUT) 
			{
			 g_SearchEval = LastEval = Eval;
			 TempBoard = g_Boardlist[0];
			 // Check if there is only one legal move, if so don't keep searching
			 if (Movelist.nMoves == 1 && g_MaxDepth > 6) Eval = TIMEOUT;
			 if ( (clock() - starttime) > (CLOCKS_PER_SEC * fMaxSeconds *.77f)  // probably won't get any useful info before timeup
				 || (abs (Eval) > 2001-SearchDepth ) ) // found a win, can stop searching now) 
				{Eval = TIMEOUT; 
				 g_SearchingMove++;} 
			}
		 else 
			{
			 if (abs(g_SearchEval) < 3000) LastEval = g_SearchEval;
			}
		}
		
	if (nDoMove == NONE) nDoMove = 1;
	if (nDoMove != NONE) 
		{
		DoMove(Movelist.Moves[nDoMove], InBoard , MAKEMOVE );	 
		g_GameMoves [ g_nMoves++ ] = Movelist.Moves[nDoMove];
		g_GameMoves [ g_nMoves ].SrcDst = 0;
		}

	RunningDisplay ( bestmove, 0 );
	return LastEval;
	}